#include <zephyr/kernel.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>
#include <bluetooth/services/nus.h>

#if defined(CONFIG_LOG)
#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(main, LOG_LEVEL_DBG);

#define CUSTOM_LOG_DBG(...) LOG_DBG(__VA_ARGS__)
#define CUSTOM_LOG_INF(...) LOG_INF(__VA_ARGS__)
#define CUSTOM_LOG_WRN(...) LOG_WRN(__VA_ARGS__)
#define CUSTOM_LOG_ERR(...) LOG_ERR(__VA_ARGS__)
#else
#define CUSTOM_LOG_DBG(...) printk("DBG: "__VA_ARGS__)
#define CUSTOM_LOG_INF(...) printk("INF: "__VA_ARGS__)
#define CUSTOM_LOG_WRN(...) printk("WRN: "__VA_ARGS__)
#define CUSTOM_LOG_ERR(...) printk("ERR: "__VA_ARGS__)
#endif

#define DEVICE_NAME                 CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN             (sizeof(DEVICE_NAME) - 1)

#define DEFAULT_SLEEP_TIME_MS       1

#define BUF_SIZE                    256

enum {
    STATE_INIT,
    STATE_START_ADVERTISING,
    STATE_WAITING_FOR_CONNECTION,
    STATE_IDLE,
    STATE_SEND_RESPONSE,
    STATE_SENDING,
    STATE_ERROR,
    STATE_DEAD,
};

/* BLE connection callbacks */
static void connected_cb(struct bt_conn *conn, uint8_t conn_err);
static void disconnected_cb(struct bt_conn *conn, uint8_t reason);
static void param_updated_cb(struct bt_conn *conn, uint16_t interval, uint16_t latency, uint16_t timeout);
static void phy_updated_cb(struct bt_conn *conn, struct bt_conn_le_phy_info *param);
static void data_len_updated_cb(struct bt_conn *conn, struct bt_conn_le_data_len_info *info);

/* NUS callbacks */
static void nus_received_cb(struct bt_conn *conn, const uint8_t *const data, uint16_t len);
static void nus_sent_cb(struct bt_conn *conn);
static void nus_send_enabled_cb(enum bt_nus_send_status status);

/* BLE buffers */
uint8_t ble_tx_buffer[BUF_SIZE];
size_t ble_tx_len;
uint8_t ble_rx_buffer[BUF_SIZE];
size_t ble_rx_len;

/* Flags for BLE events */
bool is_connected;
bool nus_subscribed;
bool ble_data_received;
bool ble_data_sent;

/* BLE advertisement */
const struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
};
const struct bt_data sd[] = {
    BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NUS_VAL),
};
struct bt_le_adv_param adv_params_connectable =
    BT_LE_ADV_PARAM_INIT(BT_LE_ADV_OPT_CONNECTABLE,
                         BT_GAP_ADV_FAST_INT_MIN_1,
                         BT_GAP_ADV_FAST_INT_MAX_1,
                         NULL);
struct bt_le_ext_adv* current_adv;
struct bt_le_ext_adv_start_param ext_adv_start_param;

/* BLE Connection */
BT_CONN_CB_DEFINE(conn_callbacks) = {
    .connected           = connected_cb,
    .disconnected        = disconnected_cb,
    .le_param_req        = NULL,
    .le_param_updated    = param_updated_cb,
    .le_phy_updated      = phy_updated_cb,
    .le_data_len_updated = data_len_updated_cb,
};
struct bt_conn* current_conn;
bt_addr_le_t peer_addr;

/* BLE NUS service */
struct bt_nus_cb nus_cb = {
    .received     = nus_received_cb,
    .sent         = nus_sent_cb,
    .send_enabled = nus_send_enabled_cb,
};

/* ------------------------ BLE connection callbacks ------------------------ */

static void connected_cb(struct bt_conn *conn, uint8_t conn_err)
{
    int res;
    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr_str, sizeof(addr_str));

    if (conn_err != 0)
    {
        CUSTOM_LOG_ERR("Failed to connect to %s (0x%02X)\n", addr_str, conn_err);
        bt_conn_unref(conn);
        current_conn = NULL;
        return;
    }

    struct bt_conn_info info;
    res = bt_conn_get_info(conn, &info);
    if (res != 0)
    {
        CUSTOM_LOG_ERR("Failed to get connection info (err %d)\n", res);
        return;
    }
    CUSTOM_LOG_INF("Connected to %s, role %hhu, interval %.2f ms, latency %hu, timeout %hu, phy %hhu\n",
                   addr_str,
                   info.role,
                   (double)(info.le.interval * 1.25f),
                   info.le.latency,
                   (info.le.timeout * 10),
                   info.le.phy->tx_phy);

    current_conn = bt_conn_ref(conn);
    if (current_conn == NULL)
    {
        CUSTOM_LOG_ERR("Failed to reference connection\n");
    }
    is_connected = true;
}

static void disconnected_cb(struct bt_conn *conn, uint8_t reason)
{
    char addr_str[BT_ADDR_LE_STR_LEN];
    bt_addr_le_to_str(bt_conn_get_dst(conn), addr_str, sizeof(addr_str));

    CUSTOM_LOG_WRN("Disconnected from %s (reason 0x%02X)\n", addr_str, reason);

    if (current_conn != NULL)
    {
        bt_conn_unref(current_conn);
        current_conn = NULL;
    }

    is_connected = false;
    nus_subscribed = false;
}

static void param_updated_cb(struct bt_conn *conn, uint16_t interval, uint16_t latency, uint16_t timeout)
{
    CUSTOM_LOG_INF("Connection parameters updated: interval %.2f ms, latency %hu, timeout %hu ms\n",
                   (double)(interval * 1.25f),
                   latency,
                   (timeout * 10));
}

static void phy_updated_cb(struct bt_conn *conn, struct bt_conn_le_phy_info *param)
{
    CUSTOM_LOG_INF("PHY updated: tx_phy = %hhu, rx_phy = %hhu\n", param->tx_phy, param->rx_phy);
}

static void data_len_updated_cb(struct bt_conn *conn, struct bt_conn_le_data_len_info *info)
{
    CUSTOM_LOG_INF("Data length updated: tx_max_len %hu, tx_max_time %hu, rx_max_len %hu, rx_max_time %hu\n",
                   info->tx_max_len,
                   info->tx_max_time,
                   info->rx_max_len,
                   info->rx_max_time);
}

/* ----------------------------- NUS callbacks ------------------------------ */

static void nus_received_cb(struct bt_conn *conn, const uint8_t *const data, uint16_t len)
{
    if (len > sizeof(ble_rx_buffer))
    {
        CUSTOM_LOG_ERR("Invalid received data length: %u\n", len);
        return;
    }

    memcpy(ble_rx_buffer, data, len);
    ble_rx_len = len;
    ble_data_received = true;
}

static void nus_sent_cb(struct bt_conn *conn)
{
    ble_data_sent = true;
}

static void nus_send_enabled_cb(enum bt_nus_send_status status)
{
    if (status == BT_NUS_SEND_STATUS_ENABLED)
    {
        CUSTOM_LOG_INF("NUS send enabled\n");
        nus_subscribed = true;
    }
    else
    {
        CUSTOM_LOG_WRN("NUS send disabled\n");
        nus_subscribed = false;
    }
}

/* ---------------------------- Main entry point ---------------------------- */

int main(void)
{
    int res;

    CUSTOM_LOG_INF("Starting Peripheral...\n");

    int state = STATE_INIT;

    for (;;)
    {
        switch (state)
        {
        case STATE_INIT:
            /* BT stack init */
            res = bt_enable(NULL);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("BT init failed (err %d)\n", res);
                return res;
            }
            CUSTOM_LOG_INF("BT initialized\n");

            /* BT NUS init */
            res = bt_nus_init(&nus_cb);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("BT NUS init failed (err %d)\n", res);
                return res;
            }
            CUSTOM_LOG_INF("BT NUS initialized\n");

            /* Advertising init */
            res = bt_le_ext_adv_create(&adv_params_connectable, NULL, &current_adv);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to create advertising set (err %d)\n", res);
                return res;
            }
            res = bt_le_ext_adv_set_data(current_adv, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to set advertising data (err %d)\n", res);
                return res;
            }
            state = STATE_START_ADVERTISING;
            break;

        case STATE_START_ADVERTISING:
            res = bt_le_ext_adv_start(current_adv, &ext_adv_start_param);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Advertising failed to start (err %d)\n", res);
                return res;
            }
            CUSTOM_LOG_INF("Advertising started\n");
            state = STATE_WAITING_FOR_CONNECTION;
            break;

        case STATE_WAITING_FOR_CONNECTION:
            if ((is_connected == true) && (current_conn != NULL))
            {
                state = STATE_IDLE;
            }
            break;

        case STATE_IDLE:
            if (ble_data_received == true)
            {
                state = STATE_SEND_RESPONSE;
            }
            break;

        case STATE_SEND_RESPONSE:
            if ((nus_subscribed == false) || (is_connected == false) && (current_conn == NULL))
            {
                CUSTOM_LOG_ERR("BLE connection lost\n");
                state = STATE_ERROR;
                break;
            }
            CUSTOM_LOG_DBG("Received % 3u bytes\n", ble_rx_len);
            ble_tx_len = ble_rx_len;
            size_t last_index = ble_tx_len - 1;
            for (size_t i = 0; i < ble_tx_len; ++i)
            {
                ble_tx_buffer[i] = ble_rx_buffer[last_index - i];
            }
            res = bt_nus_send(current_conn, ble_tx_buffer, ble_tx_len);
            if (res != 0)
            {
                CUSTOM_LOG_ERR("Failed to send acknowledgment (err %d)\n", res);
                state = STATE_ERROR;
                break;
            }
            state = STATE_SENDING;
            break;

        case STATE_SENDING:
            if (ble_data_sent == true)
            {
                ble_data_received = false;
                state = STATE_IDLE;
            }
            break;

        case STATE_ERROR:
            CUSTOM_LOG_ERR("Error state, stopping...\n");
            state = STATE_DEAD;
            break;

        case STATE_DEAD:
            break;
        }
        k_sleep(K_MSEC(DEFAULT_SLEEP_TIME_MS));
    }

    return 0;
}
